use crate::{prelude::*, update_iter};

use anyhow::Result;
use asset::RasterizerDescriptorLayouts;
use paste;
use utilities::prelude::cgmath::{Vector3, Zero};

use super::{
    super::{
        ExtensionCheck, RenderingFrontEnd,
        shared::{
            bufferhandler::*, position_buffer_reader::PositionBuffer,
            safecommandbuffer::SafeCommandBuffer,
        },
    },
    SceneInfo,
    lights::light_wrapper::{LightData, LightHandler},
    pipelines::RasterizerPipelines,
    rasterizershader::RasterizerShader,
};

use std::sync::{Arc, Mutex};

pub struct TraditionalRasterizer {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,

    user_view: View,

    // lights
    light_handler: LightHandler,

    // scene info
    scene_info_buffer: Arc<Buffer<SceneInfo>>,

    buffer_handler: BufferHandler,
    render_target: RenderTarget,

    enable_lighting: bool,
    use_shadow_maps: bool,
    sample_count: VkSampleCountFlags,
    render_scale: f32,

    position_buffer: PositionBuffer,
    position_target_index: usize,
}

impl TraditionalRasterizer {
    pub fn new(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        rasterizer_info: RasterizerInfo,
        mut graphics_info: GraphicsInfo,
    ) -> Result<Self> {
        let proj = perspective(45.0, screen_width / screen_height, 0.01, 100.0);
        let view = View::new(device, proj, 45.0, true)?;

        let info_buffer = Self::create_info_buffer(device)?;

        if graphics_info.render_scale == 0.0 {
            graphics_info.render_scale = 1.0;
        }

        let (render_target, position_target_index) = Self::create_render_target(
            device,
            queue,
            (screen_width, screen_height),
            images,
            graphics_info.render_scale,
            graphics_info.sample_count,
        )?;

        let light_handler = LightHandler::new(
            device,
            queue,
            images.mono().len(),
            true,
            rasterizer_info.shadow_image_size,
        )?;

        let buffer_handler = Self::create_buffer_handler(
            device,
            queue,
            images.mono().len(),
            render_target.render_pass().clone(),
            rasterizer_info.use_shadow_maps,
            &light_handler,
            graphics_info.sample_count,
        )?;

        Ok(Self {
            device: device.clone(),
            queue: queue.clone(),

            user_view: view,

            light_handler,

            scene_info_buffer: info_buffer,

            buffer_handler,

            position_buffer: PositionBuffer::new(
                render_target.framebuffer(0).image(position_target_index),
            ),
            position_target_index,

            render_target,

            enable_lighting: rasterizer_info.enable_lighting,
            use_shadow_maps: rasterizer_info.use_shadow_maps,
            sample_count: graphics_info.sample_count,
            render_scale: graphics_info.render_scale,
        })
    }
}

impl TraditionalRasterizer {
    fn create_render_target(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        (screen_width, screen_height): (f32, f32),
        images: &TargetMode<Vec<Arc<Image>>>,
        render_scale: f32,
        sample_count: VkSampleCountFlags,
    ) -> Result<(RenderTarget, usize)> {
        assert_eq!(
            render_scale, 1.0,
            "Traditional Rasterizer: render_scale needs an in between image to render to, then blit this into swapchain image"
        );

        let width = (screen_width * render_scale) as u32;
        let height = (screen_height * render_scale) as u32;

        let images = images.mono();

        let mut depth_target = CustomTarget::depth();
        depth_target.format = VK_FORMAT_D32_SFLOAT;

        if sample_count == VK_SAMPLE_COUNT_1_BIT {
            Ok((
                RenderTarget::builder()
                    .add_sub_pass(
                        SubPass::builder(width, height)
                            // swapchain color target
                            .set_prepared_targets(images, 0, None)
                            // position target
                            .add_target_info(CustomTarget {
                                usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT
                                    | VK_IMAGE_USAGE_TRANSFER_SRC_BIT,
                                format: VK_FORMAT_R32G32B32A32_SFLOAT,
                                clear_on_load: true,
                                store_on_save: true,
                                attach_sampler: false,
                                use_as_input: false,
                                clear_value: ClearValue::Color([
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                ]),
                            })
                            // depth target
                            .add_target_info(depth_target)
                            .set_sample_count(VK_SAMPLE_COUNT_1_BIT)
                            .use_queue(queue.clone())
                            .build(device)?,
                    )
                    .build(device)?,
                1,
            ))
        } else {
            Ok((
                RenderTarget::builder()
                    .add_sub_pass(
                        SubPass::builder(width, height)
                            // multi sampled color target
                            .add_target_info(CustomTarget {
                                usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT.into(),
                                format: images[0].vk_format(),
                                clear_on_load: true,
                                store_on_save: true,
                                attach_sampler: false,
                                use_as_input: false,
                                clear_value: ClearValue::Color([0.0, 0.0, 0.0, 0.0]),
                            })
                            // multi sampled position target
                            .add_target_info(CustomTarget {
                                usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT.into(),
                                format: VK_FORMAT_R32G32B32A32_SFLOAT,
                                clear_on_load: true,
                                store_on_save: true,
                                attach_sampler: false,
                                use_as_input: false,
                                clear_value: ClearValue::Color([
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                ]),
                            })
                            // multi sampled depth target
                            .add_target_info(depth_target)
                            .set_sample_count(sample_count)
                            // swapchain resolve targets
                            .add_resolve_targets((images, true))
                            // resolved position target
                            .add_resolve_targets(CustomTarget {
                                usage: VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT
                                    | VK_IMAGE_USAGE_TRANSFER_SRC_BIT,
                                format: VK_FORMAT_R32G32B32A32_SFLOAT,
                                clear_on_load: true,
                                store_on_save: true,
                                attach_sampler: false,
                                use_as_input: false,
                                clear_value: ClearValue::Color([
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                    std::f32::INFINITY,
                                ]),
                            })
                            .use_queue(queue.clone())
                            .build(device)?,
                    )
                    .build(device)?,
                4,
            ))
        }
    }

    fn create_buffer_handler(
        device: &Arc<Device>,
        queue: &Arc<Mutex<Queue>>,
        image_count: usize,
        render_pass: Arc<RenderPass>,
        use_shadow_maps: bool,
        light_handler: &LightHandler,
        sample_count: VkSampleCountFlags,
    ) -> Result<BufferHandler> {
        // create all descriptor set layouts, that are needed for the pipeline layout
        let globals_layout = View::desc_layout(device)?;
        let model_layout = RasterizerDescriptorLayouts::descriptor_layout(device.clone())?;
        let model_bone_layout =
            RasterizerDescriptorLayouts::bone_descriptor_layout(device.clone())?;
        let model_color_layout =
            RasterizerDescriptorLayouts::color_descriptor_layout(device.clone())?;
        let model_color_bone_layout =
            RasterizerDescriptorLayouts::color_bone_descriptor_layout(device.clone())?;
        let model_color_normal_layout =
            RasterizerDescriptorLayouts::color_normal_descriptor_layout(device.clone())?;
        let model_color_normal_bone_layout =
            RasterizerDescriptorLayouts::color_bone_normal_descriptor_layout(device.clone())?;

        let (point, directional, spot) = light_handler.light_data();

        let shader = RasterizerShader::traditional(
            device,
            use_shadow_maps,
            point.light_count,
            directional.light_count,
            spot.light_count,
        )?;

        let attachment_info = vec![true, true];

        let create_info = BufferHandlerCreateInfo::new(
            &shader,
            device,
            queue,
            image_count,
            BufferHandlerType::graphics(
                &render_pass,
                &attachment_info,
                0,
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionOnly>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionColorUV>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_color_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormal>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormalBone>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_bone_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormalColor>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormalColorUV>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_color_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormalColorUVBone>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_color_bone_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormalColorUVNormalUV>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_color_normal_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
                move |device: Arc<Device>,
                      pipeline_layout: &Arc<PipelineLayout>,
                      vertex_shader: Arc<ShaderModule<shader_type::Vertex>>,
                      fragment_shader: Arc<ShaderModule<shader_type::Fragment>>,
                      renderpass: &Arc<RenderPass>,
                      attachment_info: &[bool],
                      sub_pass_index: u32| {
                    RasterizerPipelines::multi_sample_pipeline::<PositionNormalColorUVNormalUVBone>(
                        device,
                        pipeline_layout,
                        vertex_shader,
                        fragment_shader,
                        renderpass,
                        attachment_info,
                        sub_pass_index,
                        sample_count,
                        VK_CULL_MODE_BACK_BIT,
                    )
                },
                PipelineLayoutInfo {
                    descriptor_layouts: Self::collect_layouts(
                        &model_color_normal_bone_layout,
                        &globals_layout,
                        &point,
                        &directional,
                        &spot,
                    ),
                    push_constants: &[],
                },
            ),
        );

        let mut buffer_handler = BufferHandler::new(&create_info, true)?;

        update_iter!(
            buffer_handler,
            (position, PositionOnly),
            (position_coloruv, PositionColorUV),
            (position_normal_color, PositionNormalColor),
            (position_normal_bone, PositionNormalBone),
            (position_normal, PositionNormal),
            (position_normal_coloruv, PositionNormalColorUV),
            (position_normal_coloruv_bone, PositionNormalColorUVBone),
            (
                position_normal_coloruv_normaluv,
                PositionNormalColorUVNormalUV
            ),
            (
                position_normal_coloruv_normaluv_bone,
                PositionNormalColorUVNormalUVBone
            ),
        );

        Ok(buffer_handler)
    }

    #[inline]
    fn collect_layouts<'a>(
        model: &'a Arc<DescriptorSetLayout>,
        global: &'a Arc<DescriptorSetLayout>,
        point: &'a LightData<'a>,
        directional: &'a LightData<'a>,
        spot: &'a LightData<'a>,
    ) -> Vec<&'a dyn VkHandle<VkDescriptorSetLayout>> {
        let mut layouts: Vec<&dyn VkHandle<VkDescriptorSetLayout>> =
            vec![model.as_ref(), global.as_ref()];

        if point.light_count != 0 {
            layouts.push(point.info_descriptor.as_ref());
            layouts.push(point.depth_map_descriptor.as_ref());
        }

        if directional.light_count != 0 {
            layouts.push(directional.info_descriptor.as_ref());
            layouts.push(directional.depth_map_descriptor.as_ref());
        }

        if spot.light_count != 0 {
            layouts.push(spot.info_descriptor.as_ref());
            layouts.push(spot.depth_map_descriptor.as_ref());
        }

        layouts
    }

    fn create_info_buffer(device: &Arc<Device>) -> Result<Arc<Buffer<SceneInfo>>> {
        let scene_info = SceneInfo {
            point_light_count: 0,
            directional_light_count: 0,
            spot_light_count: 0,
            view_position: Vector3::zero(),
        };

        Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&[scene_info])
            .build(device.clone())
    }

    fn recompile_shaders(&mut self, image_count: usize) -> Result<()> {
        self.buffer_handler = Self::create_buffer_handler(
            &self.device,
            &self.queue,
            image_count,
            self.render_target.render_pass().clone(),
            self.use_shadow_maps,
            &self.light_handler,
            self.sample_count,
        )?;

        Ok(())
    }

    fn invalidate(&mut self) {
        self.buffer_handler.invalidate();
        self.light_handler.invalidate();
    }
}

impl ExtensionCheck for TraditionalRasterizer {
    fn check(device: &Arc<Device>) -> bool {
        let properties = device.physical_device().properties();

        if properties.limits.maxBoundDescriptorSets < 7 {
            return false;
        }

        true
    }
}

impl RenderingFrontEnd for TraditionalRasterizer {
    fn render_scale(&self) -> f32 {
        self.render_scale
    }

    fn view_mut(&mut self) -> &mut View {
        &mut self.user_view
    }

    fn view(&self) -> &View {
        &self.user_view
    }

    fn screen_to_world(&self, x: u32, y: u32) -> Result<Option<Vector3<f32>>> {
        let position = self.position_buffer.read_pixel(x, y)?;

        Ok(
            if !position.x.is_finite() || !position.y.is_finite() || !position.z.is_finite() {
                None
            } else {
                Some(position)
            },
        )
    }

    fn set_center(&mut self, center: Vector3<f32>) -> Result<()> {
        if self.user_view.camera_mut().set_center(center) {
            self.user_view.update_buffer()?;
        }

        self.light_handler.set_center(center)?;

        // update user view
        let mut mapped = self.scene_info_buffer.map_complete()?;
        mapped[0].view_position = center;

        Ok(())
    }

    fn process(
        &mut self,
        content: Vec<&EntityObject>,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        _images: &TargetMode<Vec<Arc<Image>>>,
        indices: &TargetMode<usize>,
    ) -> Result<()> {
        // better handling (maýbe not always a rerecord is required)
        self.invalidate();

        let index = *indices.mono();

        if self.use_shadow_maps {
            self.light_handler
                .render(&content, buffer_recorder, index)?;
        }

        self.render_target.begin(
            buffer_recorder,
            VK_SUBPASS_CONTENTS_SECONDARY_COMMAND_BUFFERS,
            index,
        );

        // collect descriptor sets
        let (point, directional, spot) = self.light_handler.light_data();

        let mut sets = Vec::new();

        if point.light_count != 0 {
            sets.push(point.info_descriptor);
            sets.push(point.depth_map_descriptor);
        }

        if directional.light_count != 0 {
            sets.push(directional.info_descriptor);
            sets.push(directional.depth_map_descriptor);
        }

        if spot.light_count != 0 {
            sets.push(spot.info_descriptor);
            sets.push(spot.depth_map_descriptor);
        }

        let mut list = Vec::new();

        self.buffer_handler.execute(
            Some(self.render_target.framebuffer(index)),
            &content,
            &sets,
            &mut list,
            index,
            &self.user_view,
        )?;

        if !list.is_empty() {
            buffer_recorder.execute_commands(&list);
        }

        self.render_target.end(buffer_recorder);

        self.position_buffer.update(
            self.render_target
                .framebuffer(index)
                .image(self.position_target_index),
        )?;

        Ok(())
    }

    fn resize(
        &mut self,
        window_width: f32,
        window_height: f32,
        images: &TargetMode<Vec<Arc<Image>>>,
    ) -> Result<()> {
        let fov = 45.0;
        let proj = perspective(fov, window_width / window_height, 0.01, 100.0);

        self.user_view = View::new_proj(&self.user_view, proj, fov)?;

        let (render_target, position_target_index) = Self::create_render_target(
            &self.device,
            &self.queue,
            (window_width, window_height),
            images,
            self.render_scale,
            self.sample_count,
        )?;

        self.render_target = render_target;
        self.position_target_index = position_target_index;

        self.buffer_handler = Self::create_buffer_handler(
            &self.device,
            &self.queue,
            images.mono().len(),
            self.render_target.render_pass().clone(),
            self.use_shadow_maps,
            &self.light_handler,
            self.sample_count,
        )?;

        Ok(())
    }

    fn add_light(&mut self, light: Light) -> Result<()> {
        if self.enable_lighting {
            self.light_handler.add(light)?;

            {
                let (point, directional, spot) = self.light_handler.light_data();

                let mut mapped = self.scene_info_buffer.map_complete()?;
                mapped[0].directional_light_count = directional.light_count;
                mapped[0].point_light_count = spot.light_count;
                mapped[0].spot_light_count = point.light_count;
            }

            self.recompile_shaders(self.light_handler.image_count)?;
        }

        Ok(())
    }

    fn remove_light(&mut self, light: Light) -> Result<()> {
        if self.enable_lighting && self.light_handler.remove(&light)? {
            {
                let (point, directional, spot) = self.light_handler.light_data();

                let mut mapped = self.scene_info_buffer.map_complete()?;
                mapped[0].directional_light_count = directional.light_count;
                mapped[0].point_light_count = spot.light_count;
                mapped[0].spot_light_count = point.light_count;
            }

            self.recompile_shaders(self.light_handler.image_count)?;
        }

        Ok(())
    }

    fn clear_lights(&mut self) -> Result<()> {
        if self.enable_lighting && self.light_handler.clear()? {
            {
                let (point, directional, spot) = self.light_handler.light_data();

                let mut mapped = self.scene_info_buffer.map_complete()?;
                mapped[0].directional_light_count = directional.light_count;
                mapped[0].point_light_count = spot.light_count;
                mapped[0].spot_light_count = point.light_count;
            }

            self.recompile_shaders(self.light_handler.image_count)?;
        }

        Ok(())
    }
}
